/*
    Video: https://www.youtube.com/watch?v=oCMOYS71NIU
    Based on Neil Kolban example for IDF: https://github.com/nkolban/esp32-snippets/blob/master/cpp_utils/tests/BLE%20Tests/SampleNotify.cpp
    Ported to Arduino ESP32 by Evandro Copercini

   Create a BLE server that, once we receive a connection, will send periodic notifications.
   The service advertises itself as: 6E400001-B5A3-F393-E0A9-E50E24DCCA9E
   Has a characteristic of: 6E400002-B5A3-F393-E0A9-E50E24DCCA9E - used for receiving data with "WRITE" 
   Has a characteristic of: 6E400003-B5A3-F393-E0A9-E50E24DCCA9E - used to send data with  "NOTIFY"

   The design of creating the BLE server is:
   1. Create a BLE Server
   2. Create a BLE Service
   3. Create a BLE Characteristic on the Service
   4. Create a BLE Descriptor on the characteristic
   5. Start the service.
   6. Start advertising.

   In this example rxValue is the data received (only accessible inside that function).
   And txValue is the data to be sent, in this example just a byte incremented every second. 
*/
#include <Arduino.h>
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>

BLEServer *pServer = NULL;
BLECharacteristic * pRTxCharacteristic;
bool deviceConnected = false;
bool oldDeviceConnected = false;
uint8_t txValue = 0;

// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define SERVICE_UUID           "0000ffe0-0000-1000-8000-00805f9b34fb" // UART service UUID
#define CHARACTERISTIC_UUID_RTX "0000ffe1-0000-1000-8000-00805f9b34fb"

#define BUTTON_A_PIN 32
#define BUTTON_B_PIN 33

uint8_t params[] = { 0x66, 0x11, 0x20, 0x00, 0x64, 0x30, 0x3C, 0x2A, 0x14, 0xD9, 0x00, 0x32, 0x00, 0x64, 0x00, 0x2D, 0x00, 0xCC, 0x10, 0x64, 0x32, 0xFC, 0x53, 0x98, 0x08, 0x64, 0x00, 0x41, 0x00, 0x00, 0x00, 0x28, 0x00, 0x00, 0x00, 0x78 };
byte params_request[] = {0x66, 0x11, 0x00, 0x77};
bool send_params = false;

uint8_t status[] = { 0x66, 0x41, 0x05, 0x32, 0x32, 0x00, 0x00, 0x02 , 0x02};
byte status_request[] = {0x66, 0x41, 0x02};
bool send_status = false;

void write_status_checksum() {
  int suma = 0;
  for (int i = 0; i < sizeof(status) - 1; i++) {
    suma += status[i];
  }
  int tmp_suma = (suma / 256) * 256;
  status[8] = suma - tmp_suma;
  status[8] = 0;
}

class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

class MyCallbacks: public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic *pCharacteristic) {
      std::string rxValue = pCharacteristic->getValue();
      if (rxValue.length() > 0) {
        if( (rxValue.length() == sizeof(params_request)) && (rxValue[0] == params_request[0]) && (rxValue[1] == params_request[1]) && (rxValue[2] == params_request[2]) && (rxValue[3] == params_request[3]) ) {
          send_params = true;
        }
        else if( (rxValue.length() == (sizeof(status_request) + 3)) && (rxValue[0] == status_request[0]) && (rxValue[1] == status_request[1]) && (rxValue[2] == status_request[2]) ) {
          send_status = true;
        }
        Serial.print(rxValue.length());
        Serial.print(" ");
        Serial.print(sizeof(status_request));
        Serial.print(" ");
        Serial.print(sizeof(status));
        Serial.println();
        Serial.print("Received Value: ");
        for (int i = 0; i < rxValue.length(); i++) {
          Serial.print(rxValue[i], HEX);
          Serial.print(" ");
        }
        Serial.println();
        Serial.println("*********");
      }
    }
};


void setup() {
  Serial.begin(9600);

  pinMode(BUTTON_A_PIN, INPUT_PULLUP);
  pinMode(BUTTON_B_PIN, INPUT_PULLUP);

  // Create the BLE Device
  BLEDevice::init("RAYVOLT DUMMY");

  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pRTxCharacteristic = pService->createCharacteristic(
										CHARACTERISTIC_UUID_RTX,
										BLECharacteristic::PROPERTY_NOTIFY | BLECharacteristic::PROPERTY_WRITE
									);
                      
  pRTxCharacteristic->addDescriptor(new BLE2902());
  pRTxCharacteristic->setCallbacks(new MyCallbacks());
/*
  BLECharacteristic * pRxCharacteristic = pService->createCharacteristic(
											 CHARACTERISTIC_UUID_RX,
											BLECharacteristic::PROPERTY_WRITE
										);

  pRxCharacteristic->setCallbacks(new MyCallbacks());
*/
  // Start the service
  pService->start();

  // Start advertising
  pServer->getAdvertising()->start();
  Serial.println("Waiting a client connection to notify...");
}

void loop() {
    if (!digitalRead(BUTTON_A_PIN)) {
      Serial.println("Button A");
      if (status[3] == 0xFF) {
        status[3] = 0x00;
      }
      else {
        status[3] += 1;
      }
    }
    if (!digitalRead(BUTTON_B_PIN)) {
      Serial.println("Button B");
      if (status[3] == 0x00) {
        status[3] = 0xFF;
      }
      else {
        status[3] -= 1;
      }
    } 
    if (deviceConnected) {
      if (send_params) {
        Serial.print("sending: ");
        for (int i = 0; i < sizeof(params); i++) {
          Serial.print(params[i], HEX);
          pRTxCharacteristic->setValue(&params[i], 1);
          pRTxCharacteristic->notify();
        }
        send_params = false;
        Serial.println("");
      }
      else if (send_status) {
        write_status_checksum();
        Serial.print("sending: ");
        for (int i = 0; i < sizeof(status); i++) {
          Serial.print(status[i], HEX);
          Serial.print(" ");
          pRTxCharacteristic->setValue(&status[i], 1);
          pRTxCharacteristic->notify();
        }
        send_status = false;
        Serial.println("");
      }
        /*
        pTxCharacteristic->setValue(&txValue, 1);
        pTxCharacteristic->notify();
        txValue++;
        */
		delay(10); // bluetooth stack will go into congestion, if too many packets are sent
	}

    // disconnecting
    if (!deviceConnected && oldDeviceConnected) {
        delay(500); // give the bluetooth stack the chance to get things ready
        pServer->startAdvertising(); // restart advertising
        Serial.println("start advertising");
        oldDeviceConnected = deviceConnected;
    }
    // connecting
    if (deviceConnected && !oldDeviceConnected) {
		// do stuff here on connecting
        oldDeviceConnected = deviceConnected;
    }
}